@@ -1,91 +1,121 @@
-#$Id$
-use v5.10.1;
 package REST::Neo4p::Batch;
 use REST::Neo4p::Exceptions;
-use JSON::XS;
-use REST::Neo4p::ParseStream;
-use HOP::Stream qw/drop head/;
+use JSON::Streaming::Reader;
 require REST::Neo4p;
 
 use base qw(Exporter);
 use strict;
 use warnings;
-no warnings qw(once);
 
 BEGIN {
-  $REST::Neo4p::Batch::VERSION = '0.4000';
+  $REST::Neo4p::Batch::VERSION = '0.1';
 }
 
 our @EXPORT = qw(batch);
 our @BATCH_ACTIONS = qw(keep_objs discard_objs);
-our $BUFSIZE = 50000;
 
 sub batch (&@) {
   my ($coderef,$action) = @_;
-  my $agent = REST::Neo4p->agent;
-  if ($agent->is_version_4) {
-    REST::Neo4p::NotSuppException->throw("Batch mode not supported on Neo4j server v4.0+");
-  }
+  my $agent = $REST::Neo4p::AGENT;
   my @errors;
-  REST::Neo4p::CommException->throw("Not connected\n") unless REST::Neo4p->connected;
+  REST::Neo4p::CommException->throw('Not connected') unless $agent;
   warn 'Agent already in batch_mode on batch() call' if ($agent->batch_mode);
-  REST::Neo4p::LocalException->throw("batch requires argument 'keep_objs' or 'discard_objs'\n") unless ($action && grep(/^$action$/,qw/keep_objs discard_objs/));
   $agent->batch_mode(1);
   $coderef->();
-  my $tmpfh = $agent->execute_batch_chunk;
-  my $jsonr = JSON::XS->new->utf8;
-  my $buf;
-  $tmpfh->read($buf, $BUFSIZE);
-  $jsonr->incr_parse($buf);
-  my $res = j_parse($jsonr);
-  die "j_parse: expecting BATCH stream" unless ($res->[0] eq 'BATCH');
-  my $str = $res->[1]->();
-  while (my $obj = drop($str)) {
-    use experimental qw/smartmatch/;
-    $obj = $obj->[1];
-    given ($obj) {
-      when (!!ref($obj)) {
-	if ($obj->{status} !~ m/^2../) {
-	  warn "Error at id ".$obj->{id}." from ".$obj->{from}.": status ".$obj->{status} if $REST::Neo4p::VERBOSE;
-	  push @errors, REST::Neo4p::Neo4jException->new(
-	    code=>$obj->{status},
-	    message => 'Server returned '.$obj->{status}.' at job id '.$obj->{id}.' from '.$obj->{from}, neo4j_message=>$obj->{message}
-	   );
-	}
-	elsif (!$obj->{status}) {
-	  $obj->{status} = 599;
-	  warn "Error at id ".$obj->{id}." from ".$obj->{from}.": status ".$obj->{status} if $REST::Neo4p::VERBOSE;
-	  push @errors, REST::Neo4p::Neo4jException->new(
-	    code=>$obj->{status},
-	    message => 'Server returned no status at job id '.$obj->{id}.' from '.$obj->{from}, neo4j_message=>$obj->{message}
-	   );
-	}
-	else {
-	  _register_object($obj) if $action eq 'keep_objs';
-	}
+  for ($action) {
+    /^discard_objs$/ && do {
+      # do in eval to catch an agent error...
+      while (my $tmpf = $agent->execute_batch_chunk) {
+	@errors = _scan_for_errors($tmpf);
+	unlink $tmpf;
+	1;
       }
-      when ('PENDING') {
-	$tmpfh->read($buf,$BUFSIZE);
-	$jsonr->incr_parse($buf)
+      last;
+    };
+    /^keep_objs$/ && do {
+      while (my $tmpf = $agent->execute_batch_chunk) {
+	@errors = _scan_for_errors($tmpf);
+	_process_objs($tmpf);
       }
-      when (!defined) {
-	last;
-      }
-      default {
-	die "j_parse: batch response ended prematurely";
-      }
-    }
-
+      last;
+    };
+    do { # fallthru
+      die "I shouldn't be here."
+    };
   }
   $agent->batch_mode(undef);
   return @errors;
 }
 
+sub _scan_for_errors {
+  my $tmpf = shift;
+  open my $fh, $tmpf or REST::Neo4p::LocalException->throw("Problem with temp file $tmpf : $!");
+  my $jsonr = JSON::Streaming::Reader->for_stream($fh);
+  my $in_response;
+  my @errors;
+  PARSE :
+      while (my $cursor = $jsonr->get_token) {
+	for ($$cursor[0]) {
+	  /start_array/ && do {
+	    $in_response = 1;
+	    last;
+	  };
+	  /start_object/ && do {
+	    unless ($in_response) {
+	      REST::Neo4p::LocalException->throw("Unexpected token in server batch response");
+	    }
+	    my $obj = $jsonr->slurp;
+	    if ($obj->{status} !~ m/^2../) {
+	      warn "Error at id ".$obj->{id}." from ".$obj->{from}.": status ".$obj->{status} if $REST::Neo4p::VERBOSE;
+	      push @errors, REST::Neo4p::Neo4jException->new(code=>$obj->{status},message => 'Server returned '.$obj->{status}.' at job id '.$obj->{id}.' from '.$obj->{from}, neo4j_message=>$obj->{message});
+	    }
+	    last;
+	  };
+	  /end_array/ && do {
+	    last PARSE;
+	  };
+	}
+	1;
+      }
+  $fh->close;
+  return @errors;
+}
+
+# right now, look for 'body' elts, and
 # create new nodes, relationships as they are encountered
 #
 # TODO: handling indexes, queries? Prevent queries in batch mode?
 # TODO: use JSON streaming from file
 
+sub _process_objs {
+  my $tmpf = shift;
+  open my $fh, $tmpf or REST::Neo4p::LocalException->throw("Problem with temp file $tmpf : $!");
+  my $jsonr = JSON::Streaming::Reader->for_stream($fh);
+  my $in_response;
+  PARSE : 
+      while ( my $cursor = $jsonr->get_token ) {
+	for ($$cursor[0]) {
+	  /start_array/ && do {
+	    $in_response = 1;
+	    last;
+	  };
+	  /start_object/ && do {
+	    unless ($in_response) {
+	      REST::Neo4p::LocalException->throw("Unexpected token in server batch response");
+	    }
+	    _register_object($jsonr->slurp);
+	    last;
+	  };
+	  /end_array/ && do {
+	    last PARSE;
+	  };
+	}
+      }
+  $fh->close;
+  unlink $tmpf;
+  return;
+}
+
 sub _register_object {
   my $decoded_batch_resp = shift;
   my ($id, $from, $body) = @{$decoded_batch_resp}{qw(id from body)};
@@ -95,17 +125,14 @@ sub _register_object {
   if ($body->{template}) {
     $obj = REST::Neo4p::Index->new_from_json_response($body);
   }
-  elsif ($body->{from} and $body->{from} =~ /properties/) {
-    1; # ignore
-  }
-  elsif ($body->{self} and $body->{self} =~ m|node/[0-9]+$|) {
+  elsif ($body->{self} =~ m|node/[0-9]+$|) {
     $obj = REST::Neo4p::Node->new_from_json_response($body);
   }
-  elsif ($body->{self} and $body->{self} =~ m|relationship/[0-9]+$|) {
+  elsif ($body->{self} =~ m|relationship/[0-9]+$|) {
     $obj = REST::Neo4p::Relationship->new_from_json_response($body);
   }
   else {
-    warn "Don't understand object in batch response: id ".$id if $REST::Neo4p::VERBOSE;
+    warn "Don't understand object in batch response: id ".$id;
   }
   if ($obj) {
     my $batch_objs = $REST::Neo4p::Entity::ENTITY_TABLE->{batch_objs};
@@ -116,6 +143,10 @@ sub _register_object {
   return;
 }
 
+sub _cleanup_batch_objs {
+  
+}
+
 =head1 NAME
 
 REST::Neo4p::Batch - Mixin for batch processing
@@ -126,7 +157,7 @@ REST::Neo4p::Batch - Mixin for batch processing
  use REST::Neo4p::Batch;
  use List::MoreUtils qw(pairwise);
 
- my @bunch = map { "new_node_$_" } (1..100);
+ my @bunch = map { "new_node_$_" } (1.100);
  my @nodes;
  batch {
   my $idx = REST::Neo4p::Index->new('node','bunch');
@@ -143,15 +174,12 @@ REST::Neo4p::Batch - Mixin for batch processing
 
 =head1 DESCRIPTION
 
-REST::Neo4p::Batch adds some syntactic sugar allowing ordinary
+C<REST::Neo4p::Batch> adds some syntactic sugar allowing ordinary
 REST::Neo4p code to be processed through the Neo4j REST batch API.
 
-Batch mode is not supported in Neo4j version 4.0+. The methods in this
-module will barf.
-
 =head1 batch {} ($action)
 
-To execute server calls generated by REST::Neo4p code, 
+To execute server calls generated by C<REST::Neo4p> code, 
 wrap the code in a batch block:
 
  batch {
@@ -164,16 +192,16 @@ The C<$action> parameter B<must be> (there is no default) one of
 
 =item * 'keep_objs'
 
-If C<keep_objs> is specified, any nodes, relationships or indexes
-returned in the server reponse will be created in memory as
-REST::Neo4p objects.
+If C<keep_objs> is specified, any nodes, relationships or indexes returned 
+in the server reponse will be created in memory as C<REST::Neo4p> objects.
+This is more time-efficient if the program plans to use the objects subsequently.
 
 =item * 'discard_objs'
 
 If C<discard_objs> is specified, Neo4j entities in the server response
-will not be automatically registered as REST::Neo4p objects. Of
+will not be automatically registered as C<REST::Neo4p> objects. Of
 course, these objects can be retrieved from the server through object
-creation and other methods, outside of the batch block.
+creation and other methods, outside of the batch block. This is more space efficient if the program does not plan to use the objects subsequently: 
 
  #!perl
  # loader...
@@ -182,7 +210,7 @@ creation and other methods, outside of the batch block.
  
  open $f, shift() or die $!;
  batch {
-   while (<$f>) {
+   while (<>) {
     chomp;
     ($name, $value) = split /\t/;
     REST::Neo4p::Node->new({name => $name, value => $value});
@@ -191,10 +219,10 @@ creation and other methods, outside of the batch block.
 
 =back
 
-=head2 Errors in batch jobs
+=head2 Error in batch jobs
 
 C<batch{}()> returns returns an array of
-L<REST::Neo4p::Exceptions|REST::Neo4p::Neo4jException> error objects for each job that returns
+L<REST::Neo4p::Neo4jException> error objects for each job that returns
 a server-generated error. If no errors were encountered, it returns
 undef.
 
@@ -211,13 +239,13 @@ C<$REST::Neo4p::VERBOSE> is set.
 
 =item *
 
-No call to the server is made until after the block is executed. There
-is some magic provided, but not all object functionality is available
-to REST::Neo4p entities obtained within the C<batch> block.
+No call to the server is made until the block is executed. There is
+some magic involved, but not all object functionality is available to
+REST::Neo4p entities obtained within the C<batch> block.
 
 For example, this works:
 
- my $idx = REST::Neo4p::Index->new('node' => 'pals_of_bob');
+ my $idx = REST::Neo4p::Index->new('pals_of_bob');
  my $name = 'fred'
  batch {
   my $node = REST::Neo4p::Node->new({name => $name});
@@ -226,15 +254,15 @@ For example, this works:
 
 but this does not:
 
- my $idx = REST::Neo4p::Index->new('node' => 'pals_of_bob');
- my $name = 'fred';
+ my $idx = REST::Neo4p::Index->new('pals_of_bob');
+ my $name = 'fred'
  batch {
   my $node = REST::Neo4p::Node->new({name => $name});
   $idx->add_entry($node, name => $node->get_property('name'));
  } 'keep_objs';
 
 because $node has not been created on the server at the time that
-add_entry() is executed, so C<get_property('name')> fails.
+add_entry() is executed.
 
 =back
 
@@ -250,7 +278,7 @@ L<REST::Neo4p>, L<REST::Neo4p::Agent>
 
 =head1 LICENSE
 
-Copyright (c) 2012-2020 Mark A. Jensen. This program is free software; you
+Copyright (c) 2012 Mark A. Jensen. This program is free software; you
 can redistribute it and/or modify it under the same terms as Perl
 itself.
 
